# -*- coding: UTF-8 -*-
# Copyright 2013-2017 John Red.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import random
import math

from enigma import eTimer,eActionMap,getDesktop, ePoint # @UnresolvedImport
from Screens.Screen import Screen
from Components.Label import Label
#from Screens.Standby import inStandby
import Screens.Standby

try:
    from Plugins.Extensions.Seasondream.panels.LcdPanel import LcdPanel
    from Plugins.Extensions.Seasondream.Globals import glob
    from Plugins.Extensions.Seasondream.Logger import Logger
    
except ImportError,e:
    import traceback
    traceback.print_exc()


class ScreensaverWatcher(Logger):

    MODE_STOPPED=1
    MODE_MONITORING=2
    MODE_ACTIVE=3

    
    def __init__(self, session, showScreensaver, hideScreensaver, timeout=5*60*1000):
        Logger.__init__(self)        
        self._showScreensaver=showScreensaver
        self._hideScreensaver=hideScreensaver
        self.mode=self.MODE_MONITORING
        self.timeout=timeout
        self.session=session
        self.service_paused=False
        
        eActionMap.getInstance().bindAction("", -0x7FFFFFFF, self._keyPressed)

        self.checkTimer=eTimer()
        try:
            self.checkTimer_conn = self.checkTimer.timeout.connect(self._checkScreensaver)
        except AttributeError:
            self.checkTimer.callback.append(self._checkScreensaver)
        
        self.last_pressed=time.time()
        self.checkTimer.start(self.timeout, True)             
        self.log("Screensaver watcher started. Timeout "+str(self.timeout))

    def _checkScreensaver(self):
        self.checkTimer.stop()
        _timeSinceLastAction=(int)((time.time()-self.last_pressed)*1000)
        self.log("Time since last action "+str(_timeSinceLastAction))
        if(_timeSinceLastAction>=self.timeout and Screens.Standby.inStandby is None):
            def isServiceInUser():
                if(glob.player is not None):
                    if(glob.player.state==glob.player.STATE_PLAYING):
                        self.log("Player state playing")
                        return True
                elif(self.session.nav.getCurrentlyPlayingServiceReference() is not None and self.service_paused==False):
                    return True
                return False
            if(isServiceInUser()==False):
                self.mode=self.MODE_ACTIVE
                self._showScreensaver()
                self.log("Screensaver has been shown")
            else:
                self.log("Current servive is in use")
                self.last_pressed=time.time()
                self.checkTimer.start(self.timeout, True)  
        else:
            _tOut=(int)(self.timeout-_timeSinceLastAction)
            self.checkTimer.start(_tOut, True)  
            self.log("Next check in "+str(_tOut))
        
    def _keyPressed(self, key, flag):
        self.last_pressed=time.time()
        if(self.mode==self.MODE_ACTIVE):
            self.log("Hiding screensaver...")
            self._hideScreensaver()
            self.checkTimer.start(self.timeout, True)
            self.mode=self.MODE_MONITORING            
            return True             
        
        return False

    def pause(self):
        self.last_pressed=time.time()
        self.service_paused=True
        self.log("Pause mode enabled")
    
    def unpause(self):
        self.last_pressed=time.time()
        self.service_paused=False
        self.log("Pause mode disabled")
    
    def shutdown(self):
        self.checkTimer.stop()
        if(self.mode != self.MODE_STOPPED):
            self.mode=self.MODE_STOPPED
            try:
                eActionMap.getInstance().unbindAction("", self._keyPressed)
            except Exception,e:
                self.log("Unable to unbind action "+str(e));
        self._hideScreensaver()
        self.log("Screensaver watcher stopped")
    
    
    
class Screensaver(Screen):
    desktop = getDesktop(0)             
    w=int(desktop.size().width())
    h=int(desktop.size().height())
    skin = """
    <screen name="Screensaver" position="center,center" size="%d,%d" title="Screensaver" backgroundColor="#000000">
        <widget foregroundColor="#333333" name="text" position="0,0" size="210,31" text="Seasondream" halign="center" transparent="1" font="Sd; 30" />
    </screen>""" % (w, h)
    
    def __init__(self, session):
        Screen.__init__(self, session)
        self["text"]=Label()
        
        self.xMax=None
        self.yMax=None
        
        self.x=None
        self.y=None
        
        self.direction=(None, None)
        self.alpha=None
        
        self.timer=eTimer()
        try:
            self.timer_conn = self.timer.timeout.connect(self._onTimer)
        except AttributeError:
            self.timer.callback.append(self._onTimer)
        
        self.onLayoutFinish.append(self._onFinished)
        self.onClose.append(self._onClose)
        
    def _onTimer(self):
        if(self.y<=0 and self.y+self.direction[1]<self.y):
            #top y
            self.direction=(self.direction[0],self.direction[1]*-1.0)
        elif(self.x>=self.xMax and self.x+self.direction[0]>self.x):
            #right x
            self.direction=(self.direction[0]*-1.0,self.direction[1])
        elif(self.y>=self.yMax and self.y+self.direction[1]>self.y):
            #bottom y
            self.direction=(self.direction[0],self.direction[1]*-1.0)
        elif(self.x<=0 and self.x+self.direction[0]<self.x):
            #left x
            self.direction=(self.direction[0]*-1.0,self.direction[1])

        self.x=self.x+self.direction[0]
        self.y=self.y+self.direction[1]
        self["text"].instance.move(ePoint((int)(self.x), (int)(self.y)))
            
    
    def _onFinished(self):
        s=self["text"].instance.calculateSize()
        
        self.xMax=(self.w-s.width())*1.0
        self.yMax=(self.h-s.height())*1.0
        
        self.x=random.randint(0, self.xMax-1)*1.0
        self.y=random.randint(0, self.yMax-1)*1.0

        self["text"].instance.move(ePoint((int)(self.x), (int)(self.y)))
        
        self.alpha=random.randint(0, 360)*math.pi/180.0
        LENGTH=20.0
        dir_x=math.sqrt(LENGTH/(math.pow(math.tan(self.alpha),2)+1.0))*random.sample([1.0,-1.0],1)[0]
        dir_y=dir_x*math.tan(self.alpha)
        self.direction=(dir_x, dir_y)
        
        self.timer.start(1*200, False)
        
    def _onClose(self):
        self.timer.stop()
    
    def createSummary(self):
        return LcdPanel
