# -*- coding: UTF-8 -*-
# Copyright 2013-2017 John Red.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import urlparse
import xml.etree.ElementTree
import base64
import struct
import math
import urllib

try:
    from Plugins.Extensions.Seasondream.SeasondreamHTMLParser import SeasondreamHTMLParser
    from Plugins.Extensions.Seasondream.api.APIException import APIException
    from Plugins.Extensions.Seasondream.plugins.Plugin import Plugin
    from Plugins.Extensions.Seasondream.simplejson import loads
    from Plugins.Extensions.Seasondream.plugins.Plugin import boundFunction
    from Plugins.Extensions.Seasondream.Config import cfg
    
except ImportError:
    try:
        from resources.lib.plugins.Plugin import boundFunction
        from resources.lib.utils.SeasondreamHTMLParser import SeasondreamHTMLParser
        from resources.lib.api.APIException import APIException
        from resources.lib.plugins.Plugin import Plugin
        from resources.lib.simplejson import loads
    except ImportError:
        from SeasondreamHTMLParser import SeasondreamHTMLParser
        from api.APIException import APIException
        from plugins.Plugin import Plugin
        from simplejson import loads

class Kinoylei(Plugin):

    PROVIDER_ID=1
    def __init__(self, api):
        Plugin.__init__(self, api)

    def readBoxHeader(self, data, pos=0):
        boxSize, = struct.unpack_from(">L", data, pos)
        boxType = data[pos + 4 : pos + 8]
        if boxSize == 1:
            boxSize, = struct.unpack_from(">Q", data, pos + 8)
            boxSize -= 16
            pos += 16
        else:
            boxSize -= 8
            pos += 8
        if boxSize <= 0:
            boxSize = 0
        return (pos, boxType, boxSize)

    def readInt8(self, data, pos):
        return struct.unpack_from(">B", data, pos)[0]
    
    def readByte(self, data, pos):
        return struct.unpack_from(">B", data, pos)[0]
    
    def readInt32(self, data, pos):
        return struct.unpack_from(">L", data, pos)[0]

    def readInt24(self, data, pos):
        return struct.unpack_from(">L", "\0" + data[pos:pos + 3], 0)[0]

    
    def readInt64(self, data, pos):
        a, b     = struct.unpack_from('>LL', data, pos)
        return (a << 32) | b

    def readString(self,data, pos):
        _len=0
        while (data[pos + _len] != "\0"):
            _len+=1
            
        _str=struct.unpack_from(">"+str(_len)+"s", data, pos)[0]
        pos+=_len+1
        return (_str,pos)
    
    def parseAfrtBox(self,afrt, pos):

        fragTable         = []
        version           = self.readByte(afrt, pos)
        flags             = self.readInt24(afrt, pos + 1)
        timescale         = self.readInt32(afrt, pos + 4)
        qualityEntryCount = self.readByte(afrt, pos + 8)
        pos += 9
        qualitySegmentUrlModifiers=[]
        for _ in range(0,qualityEntryCount):
            _str,pos=self.readString(afrt, pos)
            qualitySegmentUrlModifiers.append(_str)
        
        fragEntries = self.readInt32(afrt, pos)
        pos += 4
        for _ in range(fragEntries):
            firstFragment=self.readInt32(afrt, pos)
            firstFragmentTimestamp=self.readInt64(afrt, pos + 4)
            fragmentDuration=self.readInt32(afrt, pos + 12)
            discontinuityIndicator=None
            pos += 16;
            if (fragmentDuration == 0):
                discontinuityIndicator = self.readByte(afrt, pos)
                pos+=1
            fragTable.append({"firstFragment":firstFragment,"firstFragmentTimestamp":firstFragmentTimestamp,"fragmentDuration":fragmentDuration,"discontinuityIndicator":discontinuityIndicator})
            
        return fragTable;
    
    def parseAsrtBox(self,asrt, pos):
        segTable          = []
        version           = self.readByte(asrt, pos)
        flags             = self.readInt24(asrt, pos + 1)
        qualityEntryCount = self.readByte(asrt, pos + 4)
        pos += 5
        qualitySegmentUrlModifiers=[]
        for _ in range(0,qualityEntryCount):
            _str,pos=self.readString(asrt, pos)
            qualitySegmentUrlModifiers.append(_str)
        segCount = self.readInt32(asrt, pos)
        pos += 4
        for _ in range(0,segCount):
            firstSegment = self.readInt32(asrt, pos);
            fragmentsPerSegment=self.readInt32(asrt, pos + 4);
            if (fragmentsPerSegment & 0x80000000!=0):
                fragmentsPerSegment = 0;
            segTable.append({"firstSegment":firstSegment,"fragmentsPerSegment":fragmentsPerSegment})
            pos+=8
        return segTable
            
    def parseBootstrapBox(self,data, pos):
        version = self.readByte(data, pos)
        flags = self.readInt24(data, pos + 1)
        bootstrapVersion = self.readInt32(data,pos + 4)
        byte             = self.readByte(data, pos + 8)
        profile          = (byte & 0xC0) >> 6
        update = (byte & 0x10) >> 4;

        timescale           = self.readInt32(data, pos + 9)
        currentMediaTime    = self.readInt64(data, pos + 13)
        smpteTimeCodeOffset = self.readInt64(data, pos + 21)
        pos += 29;
        movieIdentifier, pos  = self.readString(data, pos)
        
        serverEntryCount = self.readByte(data, pos)
        pos+=1
        
        serverEntryTable=[]
        for i in range(0,serverEntryCount):
            set,pos=self.readString(data, pos)
            serverEntryTable.append(set)

        qualityEntryCount = self.readByte(data, pos);
        pos+=1
        qualityEntryTable=[]
        for i in range(0,qualityEntryCount):
            qet,pos=self.readString(data, pos)
            qualityEntryTable.append(qet)

        drmData,pos          = self.readString(data, pos)
        metadata,pos         = self.readString(data, pos)
        segRunTableCount     = self.readByte(data, pos)
        pos+=1
        segTable=None
        for i in range(0,segRunTableCount):
            pos, boxType, boxSize=self.readBoxHeader(data=data, pos=pos)
            if(boxType=="asrt"):
                segTable=self.parseAsrtBox(data,pos)
            pos=pos+boxSize

        fragRunTableCount = self.readByte(data, pos)
        pos+=1
        fragTable=None
        for _ in range(0,fragRunTableCount):
            pos, boxType, boxSize=self.readBoxHeader(data=data, pos=pos)
            if(boxType=="afrt"):
                fragTable=self.parseAfrtBox(data,pos)
            pos=pos+boxSize
        return segTable, fragTable           
    
    def parseManifest(self, _xml):
        root=xml.etree.ElementTree.fromstring(_xml)
        
        _media_data = {}
        _media_data["duration"] = float(root.find("{http://ns.adobe.com/f4m/1.0}duration").text)
        baseURL=root.find("{http://ns.adobe.com/f4m/1.0}baseURL")
        if(baseURL is not None):
            _media_data["baseURL"]=baseURL.text
        bitrate_list=[]
        for media in root.findall("{http://ns.adobe.com/f4m/1.0}media"):
            if("bitrate" in media.attrib):
                bitrate_list.append(int(media.attrib["bitrate"]))
        
        bitrate_list.sort()
        self.log("Available bitrates "+str(bitrate_list))
        self.log("Desired quality "+str(cfg.vkQuality.getValue()))
        _bitrate={"240":0,"360":1,"480":2,"720":3}[cfg.vkQuality.getValue()]

        _factor=len(bitrate_list)*1.0/4.0*1.0 #4 quality types
        _pos=min(len(bitrate_list)-1, (int(math.ceil(_bitrate*_factor))))
        self.log("Ceil "+str(math.ceil(_bitrate*_factor))+", factor "+str(_factor)+", take br: "+str(bitrate_list[_pos]))

        #Parse Media
        for media in root.findall("{http://ns.adobe.com/f4m/1.0}media"):
            if("bitrate" in media.attrib and int(media.attrib["bitrate"])==bitrate_list[_pos]):
                pass
            else:
                continue
            if("bitrate" in media.attrib):
                _media_data["bitrate"]=int(media.attrib["bitrate"])
            if("bootstrapInfoId" in media.attrib):
                _media_data["bootstrapInfoId"] = media.attrib["bootstrapInfoId"]
            if("url" in media.attrib):
                _media_data["url"] = media.attrib["url"]
            if("streamId" in media.attrib):
                _media_data["streamId"] = media.attrib["streamId"]
            #TODO: select bitrate
            _media_data["metadata"] = base64.b64decode(media.find("{http://ns.adobe.com/f4m/1.0}metadata").text)
            self.log("Found media "+_media_data["streamId"]+" bitrate "+str(_media_data["bitrate"]))
            break

        for bootstrapInfo in root.findall("{http://ns.adobe.com/f4m/1.0}bootstrapInfo"):
            if("id" in bootstrapInfo.attrib and bootstrapInfo.attrib["id"]==_media_data["bootstrapInfoId"]):
                _media_data["bootstrapInfo"]=base64.b64decode(bootstrapInfo.text)
                break

        pos, boxType, _ =self.readBoxHeader(data=_media_data["bootstrapInfo"])
        if (boxType != "abst"):
            raise APIException(-2001,"abst box not found")
        
        _media_data["segTable"], _media_data["fragTable"] = self.parseBootstrapBox(data=_media_data["bootstrapInfo"], pos=pos)
        if(_media_data["segTable"] is None or _media_data["fragTable"] is None):
            raise APIException(-2002,"segTable or fragTable not found")
            
        return _media_data
    
    def _future(self, url):
        if(cfg.VERSION_MAJOR<2.2):
            raise APIException(-2000,"Для просмотра необходима версия 2.2 или выше.")
        self.log("Loading future object "+str(url))
        try:
            #parser = HDKinomirParser()
            iframe_src=self.read(url)
            #parser.close()
            _ret=self.parseIframe(url,iframe_src)
            if(_ret is not None):
                return _ret[0]
        except Exception:
            pass
        return None
    
    def _getAvailableSources(self, url, src):
        import re
        ret={}
        for item in re.compile('{(.[^{}]*?)}').findall(src.replace("\n","").replace("\r","")):
            if(item.find("video_token")>-1 and item.find("d_id")>-1):
                self.log("session params "+ item)
                data=dict();
                for item2 in re.compile('(.[^:]*?):(.[^,]*?)(,|$)').findall(item):
                    key=item2[0].strip()
                    value=item2[1].strip().replace("'","").replace("\"","").replace("null","")
                    if("cd"==key):
                        continue
                    data[key]=value
                data["cd"]=0
                _parsed_url=urlparse.urlparse(url)
                ret=loads(self.read(_parsed_url.scheme+"://"+_parsed_url.netloc+"/sessions/create_session", options={"X-Requested-With":"XMLHttpRequest"},data=urllib.urlencode(data)),"utf-8")
                return ret
        return None  
            
    def parseIframe(self, url, src):
        ret=[]
        if(url is not None):
            _available_sources=self._getAvailableSources(url,src)
            self.log("Available sources "+str(_available_sources))
            if("manifest_f4m" in _available_sources):
                manifest_url=_available_sources["manifest_f4m"]
            else:
                self.log("manifest_f4m not found")
                return None
            _media_data=self.parseManifest(self.read(manifest_url))
            pair={}
            pair["comment"]=u"Фильм "+str(_media_data["bitrate"])+"kbps"
            pair["file"]="movie.flv"
            pair["seen"]=False
            pair["type"]="F4M"
            pair["media_data"]=_media_data
            ret.append(pair)
            return ret
        else:
            self.log("Video token not available")
        return None
    
    def _getPlaylist(self, url):
        #if(cfg.VERSION_MAJOR<2.2):
        #    raise APIException(-2000,"Для просмотра необходима версия 2.2 или выше.")
        src=self.read("http://moonwalk.co/player_api?w=607&h=315&fixvk=1", options={"Referer":url})
        
        parser = KinoyleiParser()
        parser.feed(self.read(url))
        parser.close()
        for src in parser.iframe:
            _query = self._parse_qs(urlparse.urlparse(src).query)
            if("url" not in _query):
                pass
            else:
                src=_query["url"]
                
            src="http://moonwalk.cc/video/30786f05e70583af/iframe"
            self.log("Parsing iframe "+src)
            parser = KinoyleiParser()
            iframe_src=self.read(src)
            parser.feed(iframe_src)
            parser.close()
            
            if(parser.selectedEpisode is not None and parser.selectedSeason is not None and len(parser.episode)>0):
                _parsed_url=urlparse.urlparse(src)
                self.log("Found #"+str(len(parser.episode))+" episodes. selected season "+str(parser.selectedSeason))
                ret=[]
                for episode in parser.episode:
                    #scheme='http', netloc='serpens.nl', path='/serial/bd1f50a721eedfbd0a8bb8c6049a6d87/iframe'
                    _file=_parsed_url.scheme+"://"+_parsed_url.netloc+_parsed_url.path+"?"+urllib.urlencode({"season":parser.selectedSeason,"episode":episode,"referer":url})
                    self.log("URL "+str(_file))
                    pair={}
                    pair["future_media_data"]=boundFunction(self._future, _file)
                    pair["comment"]=u"Серия "+str(episode)
                    pair["file"]="movie.flv"
                    pair["seen"]=False
                    pair["type"]="F4M"
                    pair["media_data"]=None
                    ret.append(pair)
                return ret
            else:
                _ret=self.parseIframe(src, iframe_src)
                if(_ret is not None):
                    return _ret
        raise APIException(-2002,Plugin.ERROR_MESSAGE_INFO)



class KinoyleiParser(SeasondreamHTMLParser):

    def __init__(self):
        SeasondreamHTMLParser.__init__(self)
        self.iframe = []
        self.title=None
        self.title_data=False
        self.video_token=None
        self.selectEpisode=False
        self.selectSeason=False
        self.episode=[]
        self.selectedEpisode=None
        self.season=[]
        self.selectedSeason=None
        
    def unescape(self, s):
        return  s.replace("&amp;","&")

    def handle_data(self, data):
        if(self.title_data):
            self.title=data
            if(self.title.startswith("video ")):
                self.video_token=self.title[6:]

    def handle_starttag(self, tag, attrs):
        if(tag.lower()=="iframe"):
            src = ""
            try:
                src=[v for k, v in attrs if k == 'src'][0]
            except Exception, e:
                pass
            if(src.find("iframe")>-1):
                self.iframe.append(src)
        elif(tag.lower()=="title"):
            self.title_data=True
        elif(tag.lower()=="select"):
            id = ""
            try:
                id=[v for k, v in attrs if k == 'id'][0]
            except Exception, e:
                pass
            if(id=="episode"):
                self.selectEpisode=True
            elif(id=="season"):
                self.selectSeason=True
        elif(tag.lower()=="option" and self.selectEpisode==True):
            value = ""
            try:
                value=[v for k, v in attrs if k == 'value'][0]
            except Exception, e:
                pass
            try:
                _selected=[v for k, v in attrs if k == 'selected'][0]
                if(_selected.lower()=="selected"):
                    self.selectedEpisode=value
            except Exception, e:
                pass
            self.episode.append(value)
        elif(tag.lower()=="option" and self.selectSeason==True):
            value = ""
            try:
                value=[v for k, v in attrs if k == 'value'][0]
            except Exception, e:
                pass
            try:
                _selected=[v for k, v in attrs if k == 'selected'][0]
                if(_selected.lower()=="selected"):
                    self.selectedSeason=value
            except Exception, e:
                pass
            self.season.append(value)
            
    def handle_endtag(self, tag):
        if(tag.lower()=="title"):
            self.title_data=False
        elif(tag.lower()=="select"):
            self.selectEpisode=False
            self.selectSeason=False
